import json

import boto3

from .aws_base import AwsBase


class Topic(AwsBase):
    def __init__(self, topic_name, region_name="eu-west-1"):
        super().__init__("sns", region_name)
        self._topic_arn = (
            boto3.resource(self._service_name, region_name)
            .create_topic(Name=topic_name)
            .arn
        )

    async def publish(self, subject, message):
        await self.init_client_if_required()
        await self._client.publish(
            TargetArn=self._topic_arn,
            Message=json.dumps({"default": json.dumps(message)}),
            Subject=subject,
            MessageStructure="json",
        )



import asyncio
import logging
from contextlib import AsyncExitStack

from aiobotocore.session import get_session, AioSession
from aiobotocore.config import AioConfig

_LOGGER = logging.getLogger(__name__)


def get_aiosession() -> AioSession:
    return get_session()


class AwsBase:
    def __init__(self, service_name, region_name="eu-west-1", max_pool_connections=10) -> None:
        self._client = None
        self._context_stack = AsyncExitStack()
        self._region_name = region_name
        self._service_name = service_name
        self._init_lock = asyncio.Lock()
        self._max_pool_connections = max_pool_connections

    async def close(self) -> None:
        await self._context_stack.aclose()
        self._client = None

    async def init_client_if_required(self, session=None) -> None:
        async with self._init_lock:
            if self._client is not None:
                return

            config = AioConfig(
                max_pool_connections=self._max_pool_connections,
            )

            session = session or get_aiosession()
            self._client = await self._context_stack.enter_async_context(
                session.create_client(self._service_name, region_name=self._region_name, config=config)
            )



"""SQS Queue object to subscribe queue to topic, produce and consume messages.

#TODO: Shouldn't have to subscribe to an SNS topic to get messages from SQS.

"""
import json
from typing import List, Optional
import logging
import time

from .aws_base import AwsBase, get_aiosession, AioSession
from types_aiobotocore_sqs.client import SQSClient
from types_aiobotocore_sqs.type_defs import SendMessageResultTypeDef

_LOGGER = logging.getLogger(__name__)


class MessageHandle:
    def __init__(self, msg):
        self._msg = msg

    @property
    def body(self):
        return self._msg["Body"]

    @property
    def receipt_handle(self):
        return self._msg["ReceiptHandle"]


class Queue(AwsBase):
    """Queue object to publish and consume messages from AWS SQS, async."""

    _client: SQSClient

    def __init__(self, queue_name, region_name="eu-west-1") -> None:
        self._queue_name = queue_name
        self.queue_url: Optional[str] = None
        super().__init__("sqs", region_name)

    async def subscribe_topic(self, topic_name: str, session: AioSession = None) -> None:
        """Subscribe `Queue` to a topic.

        :param topic_name: name of the topic to subscribe to
        :type topic_name: str
        :param session: the aiobotocore session to use, defaults to None
        :type session: AioSession, optional
        """
        session = session or get_aiosession()
        await self.init_client_if_required(session)

        response = await self._client.create_queue(QueueName=self._queue_name)
        self.queue_url = response["QueueUrl"]
        attr_response = await self._client.get_queue_attributes(QueueUrl=self.queue_url, AttributeNames=["All"])

        queue_attributes = attr_response.get("Attributes")
        queue_arn = queue_attributes.get("QueueArn")

        if "Policy" in queue_attributes:
            policy = json.loads(queue_attributes["Policy"])
        else:
            policy = {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "Sid" + str(int(time.time())),
                        "Effect": "Allow",
                        "Principal": {"AWS": "*"},
                        "Action": ["sqs:SendMessage", "sqs:ReceiveMessage"],
                    }
                ],
            }

        statement = policy.get("Statement", [{}])[0]
        statement["Resource"] = statement.get("Resource", queue_arn)
        statement["Condition"] = statement.get("Condition", {})

        statement["Condition"]["StringLike"] = statement["Condition"].get("StringLike", {})
        source_arn = statement["Condition"]["StringLike"].get("aws:SourceArn", [])
        if not isinstance(source_arn, list):
            source_arn = [source_arn]

        sns = await self._context_stack.enter_async_context(session.create_client("sns", region_name=self._region_name))
        response = await sns.create_topic(Name=topic_name)
        topic_arn = response["TopicArn"]

        if topic_arn not in source_arn:
            source_arn.append(topic_arn)
            statement["Condition"]["StringLike"]["aws:SourceArn"] = source_arn
            policy["Statement"] = statement
            await self._client.set_queue_attributes(QueueUrl=self.queue_url, Attributes={"Policy": json.dumps(policy)})

        await sns.subscribe(TopicArn=topic_arn, Protocol=self._service_name, Endpoint=queue_arn)
        await sns.close()

    async def receive_message(self, num_msgs: int = 1) -> List[MessageHandle]:
        """
        Consume messages from the queue.

        :param num_msgs: max number of messages to receive. Default 1
        :type num_msgs: int
        :return: Message in the form of a message handle
        :rtype: list[MessageHandle]
        """
        if self.queue_url is None:
            _LOGGER.error("No subscribed queue")
            return [None]
        response = await self._client.receive_message(QueueUrl=self.queue_url, MaxNumberOfMessages=num_msgs)
        res = []
        for msg in response.get("Messages", []):
            res.append(MessageHandle(msg))
        return res

    async def delete_message(self, msg_handle: MessageHandle) -> None:
        """Delete a message from the queue.

        :param msg_handle: Message handle to delete, will delete based on `msg_handle.receipt_handle`
        :type msg_handle: MessageHandle
        """
        await self._client.delete_message(QueueUrl=self.queue_url, ReceiptHandle=msg_handle.receipt_handle)

    async def send_message(self, message: dict, **kwargs) -> Optional[SendMessageResultTypeDef]:
        """Publish a message to the queue.

        Example of returned metadata dict:
        {'MD5OfMessageBody': '83956d1d4f05f535d55c9895ff593550',
            'MessageId': 'd9654e21-f92b-4d27-bcb5-b27e1dc18839',
            'ResponseMetadata': {'HTTPHeaders': {'content-length': '378',
                                                'content-type': 'text/xml',
                                      'date': 'Tue, 20 Sep 2022 07:00:38 GMT',
                                      'x-amzn-requestid': '7202509e-f09f-5082-8aea-0b3d97c92220'},
                      'HTTPStatusCode': 200,
                      'RequestId': '7202509e-f09f-5082-8aea-0b3d97c92220',
                      'RetryAttempts': 0}}

        :param message: Message to publish, must be serializable to JSON
        :type message: dict
        :return: sent message metadata
        :rtype: SendMessageResultTypeDef
        """
        if self.queue_url is None:
            _LOGGER.error("No subscribed queue, call subscribe_topic first")
            return
        await self.init_client_if_required()
        resp: SendMessageResultTypeDef = await self._client.send_message(
            QueueUrl=self.queue_url, MessageBody=json.dumps(message), **kwargs
        )
        return resp



from setuptools import setup

install_requires = [
    "aiobotocore",
    "boto3",
    "botocore",
    "aiohttp",
    "async_timeout",
    "types-aiobotocore-sqs"
]


setup(
    name="tibber_aws",
    packages=["tibber_aws"],
    install_requires=install_requires,
    version="0.15.0",
    description="A python3 library to communicate with Aws",
    python_requires=">=3.7.0",
    author="Tibber",
    author_email="drdaniel@tibber.com",
    url="https://github.com/tibber/tibber-pyAws",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import datetime
import logging
import zlib

import botocore

from .aws_base import AwsBase

_LOGGER = logging.getLogger(__name__)

STATE_NOT_EXISTING = "not_existing"
STATE_OK = "ok"
STATE_PRECONDITION_FAILED = "precondition_failed"


class S3Bucket(AwsBase):
    def __init__(self, bucket_name, region_name="eu-west-1", **kwargs):
        self._bucket_name = bucket_name
        super().__init__("s3", region_name, **kwargs)

    async def load_metadata(self, key):
        await self.init_client_if_required()
        try:
            meta = await self._client.head_object(
                Bucket=self._bucket_name,
                Key=key,
            )
            return meta, STATE_OK
        except botocore.exceptions.ClientError as exp:
            if "Not Found" in str(exp):
                return None, STATE_NOT_EXISTING
            raise

    async def load_data_metadata(self, key, if_unmodified_since=None):
        await self.init_client_if_required()
        if if_unmodified_since is None:
            if_unmodified_since = datetime.datetime(1900, 1, 1)
        try:
            raw = await self._client.get_object(
                Bucket=self._bucket_name,
                Key=key,
                IfUnmodifiedSince=if_unmodified_since,
            )
            meta = raw["ResponseMetadata"]
            res = await raw["Body"].read()
        except self._client.exceptions.NoSuchKey:
            return None, None, STATE_NOT_EXISTING
        except botocore.exceptions.ClientError as exp:
            if "PreconditionFailed" in str(exp):
                return None, None, STATE_PRECONDITION_FAILED
            raise

        if len(key) > 3 and key[-3:] == ".gz":
            return (
                zlib.decompressobj(zlib.MAX_WBITS | 16).decompress(res),
                meta,
                STATE_OK,
            )
        return res.decode("utf-8"), meta, STATE_OK

    async def load_data(self, key, if_unmodified_since=None):
        data, _, state = await self.load_data_metadata(key, if_unmodified_since)
        return data, state

    async def delete_file(self, key):
        await self.init_client_if_required()
        await self._client.delete_object(Bucket=self._bucket_name, Key=key)

    async def store_data(self, key, data, retry=1):
        await self.init_client_if_required()

        if len(key) > 3 and key[-3:] == ".gz":
            compressor = zlib.compressobj(wbits=zlib.MAX_WBITS | 16)
            body = compressor.compress(data)
            body += compressor.flush()
        else:
            body = data

        try:
            return await self._client.put_object(
                Bucket=self._bucket_name, Key=key, Body=body
            )
        except self._client.exceptions.NoSuchBucket:
            if retry <= 0:
                raise
            await self.create()
            return await self.store_data(key, data, retry - 1)

    async def create(self):
        await self.init_client_if_required()
        try:
            await self._client.create_bucket(
                Bucket=self._bucket_name,
                CreateBucketConfiguration={"LocationConstraint": self._region_name},
            )
        except self._client.exceptions.BucketAlreadyOwnedByYou:
            _LOGGER.warning("Bucket %s already exists", self._bucket_name)

    async def list_keys(self, prefix=""):
        """Lists ALL objects of the bucket in the given prefix.
        Args:
            :prefix (str, optional): a prefix of the bucket to list (Default: none)
        Returns:
            list: The list of objects::
                [
                    {
                        'Key': 'prefix/file.json',
                        'LastModified': datetime.datetime(2018, 12, 13, 14, 15, 16, tzinfo=tzutc()),
                        'ETag': '"58bcd9641b1176ea012b6377eb5ce050"'
                        'Size': 262756,
                        'StorageClass': 'STANDARD'
                    }
                ]
        """
        await self.init_client_if_required()

        paginator = self._client.get_paginator("list_objects_v2")
        objects = []
        try:
            async for resp in paginator.paginate(
                Bucket=self._bucket_name, Prefix=prefix
            ):
                objects.extend(resp.get("Contents", []))
        except self._client.exceptions.NoSuchBucket:
            return []
        return objects


class VersionedS3Bucket(S3Bucket):
    def __init__(
        self, bucket_name, expiration_days, prefix, region_name="eu-west-1", **kwargs
    ):
        self._expiration_days = expiration_days
        self._prefix = prefix
        super().__init__(bucket_name, region_name, **kwargs)

    async def create(self):
        await super().create()
        await self._client.put_bucket_versioning(
            Bucket=self._bucket_name, VersioningConfiguration={"Status": "Enabled"}
        )
        await self._client.put_bucket_lifecycle_configuration(
            Bucket=self._bucket_name,
            LifecycleConfiguration={
                "Rules": [
                    {
                        "Status": "Enabled",
                        "Prefix": self._prefix,
                        "NoncurrentVersionExpiration": {
                            "NoncurrentDays": self._expiration_days
                        },
                    }
                ]
            },
        )
